import re
from typing import Any


def match(dict_values):
    """
    Creates one or more ``EQUAL()`` expressions for each provided dict value.
    If more than one assetions is included, the expressions are
    groupped together into using ``AND()``.

    This function also handles escaping field names and casting python values
    to the appropriate airtable types using :func:`to_airtable_value` on all
    provided values to help generate the expected formula syntax.

    Args:
        dict_values: dictionary containing column names and values

    Usage:
        >>> match({"First Name": "John", "Age": 21})
        "AND({First Name}='John',{Age}=21)"
        >>> match({"First Name": "John"})
        "{First Name}='John'"
        >>> match({"Registered": True})
        "{Registered}=1"
        >>> match({"Owner's Name": "Mike"})
        "{Owner\'s Name}='Mike'"

    """
    expressions = []
    for key, value in dict_values.items():
        expression = EQUAL(FIELD(key), to_airtable_value(value))
        expressions.append(expression)

    if len(expressions) == 0:
        return ""
    elif len(expressions) == 1:
        return expressions[0]
    else:
        return AND(*expressions)


def escape_quotes(value: str):
    r"""
    Ensures any quotes are escaped. Already escaped quotes are ignored.

    Args:
        value: text to be escaped

    Usage:
        >>> escape_quotes("Player's Name")
        Player\'s Name
        >>> escape_quotes("Player\'s Name")
        Player\'s Name
    """
    escaped_value = re.sub("(?<!\\\\)'", "\\'", value)
    return escaped_value


def to_airtable_value(value: Any):
    """
    Cast value to appropriate airtable types and format.
    For example, to check ``bool`` values in formulas, you actually to compare
    to 0 and 1.

    Arg:
        value: value to be cast.

    * ``bool`` -> ``int``
    * ``str`` -> Text is wrapped in `'single quotes'`. Existing quotes are escaped.
    * ``float``, ``int`` -> no change
    """
    if isinstance(value, bool):
        return int(value)
    elif isinstance(value, (int, float)):
        return value
    elif isinstance(value, str):
        return STR_VALUE(escape_quotes(value))
    else:
        return value


def EQUAL(left: Any, right: Any) -> str:
    """
    Creates an equality assertion

    >>> EQUAL(2,2)
    '2=2'
    """
    return "{}={}".format(left, right)


def FIELD(name: str) -> str:
    """
    Creates a reference to a field

    Args:
        name: field name

    Usage:
        >>> FIELD("First Name")
        '{First Name}'
    """
    return "{%s}" % name


def STR_VALUE(value: str) -> str:
    return "'{}'".format(value)


def IF(logical, value1, value2) -> str:
    """
    Creates an IF statement

    >>> IF("1=1"", 0, 1)
    'IF("1=1"", 0, 1)'
    """
    return "IF({}, {}, {})".format(logical, value1, value2)


def AND(*args) -> str:
    """
    Creates an AND Statement

    >>> AND(1, 2, 3)
    'AND(1, 2, 3)'
    """
    return "AND({})".format(",".join(args))



"""
The :class:`Model` class allows you create an orm-style class for your
Airtable tables.

>>> from pyairtable.orm import Model, fields
>>> class Contact(Model):
...     first_name = fields.TextField("First Name")
...     last_name = fields.TextField("Last Name")
...     email = fields.EmailField("Email")
...     is_registered = fields.CheckboxField("Registered")
...     partner = fields.LinkField("Partner", "Contact", lazy=False)
...
...     class Meta:
...         base_id = "appaPqizdsNHDvlEm"
...         table_name = "Contact"
...         api_key = "keyapikey"


Once you have a class, you can create new objects to represent your
Airtable records. Call :meth:`~pyairtable.orm.model.Model.save` to create a new record.

>>> contact = Contact(
...     first_name="Mike",
...     last_name="McDonalds",
...     email="mike@mcd.com",
...     is_registered=False
... )
...
>>> assert contact.id is None
>>> contact.exists()
False
>>> assert contact.save()
>>> contact.exists()
True
>>> contact.id
rec123asa23


You can read and modify attributes. If record already exists,
:meth:`~pyairtable.orm.model.Model.save` will update the record:

>>> assert contact.is_registered is False
>>> contact.is_registered = True
>>> contact.save()
>>> assert contact.is_registered = True
>>> contact.to_record()
{
    "id": recS6qSLw0OCA6Xul",
    "createdTime": "2021-07-14T06:42:37.000Z",
    "fields": {
        "First Name": "Mike",
        "Last Name": "McDonalds",
        "Email": "mike@mcd.com",
        "Resgistered": True
    }
}

Finally, you can use :meth:`~pyairtable.orm.model.Model.delete` to delete the record:

>>> contact.delete()
True

"""
import abc
from pyairtable import Table
from typing import TypeVar, Type, List

from .fields import Field

T = TypeVar("T", bound="Model")


class Model(metaclass=abc.ABCMeta):
    """
    This class allows you create an orm-style class for your Airtable tables.

    This is a meta class and can only be used to define sub-classes.

    The ``Meta`` is reuired and must specify all three attributes: ``base_id``,
    ``table_id``, and ``api_key``.

    >>> from pyairtable.orm import Model, fields
    >>> class Contact(Model):
    ...     first_name = fields.TextField("First Name")
    ...     age = fields.IntegerField("Age")
    ...
    ...     class Meta:
    ...         base_id = "appaPqizdsNHDvlEm"
    ...         table_name = "Contact"
    ...         api_key = "keyapikey"
    ...         timeout: Optional[Tuple[int, int]] = (5, 5)
    ...         typecast: bool = True
    """

    id: str = ""
    created_time: str = ""
    _table: Table
    _fields: dict = {}
    _linked_cache: dict = {}

    def __init_subclass__(cls, **kwargs):
        cls._validate_class()
        super().__init_subclass__(**kwargs)

    @classmethod
    def _attribute_descriptor_map(cls):
        """
        Returns a dictionary mapping the model's attribute names to the field's

        >>> class Test(Model):
        ...     first_name = TextField("First Name")
        ...     age = NumberField("Age")
        ...
        >>> Test._attribute_descriptor_map()
        >>> {
        ...     "field_name": <TextField field_name="First Name">,
        ...     "another_Field": <NumberField field_name="Age">,
        ... }
        """
        return {k: v for k, v in cls.__dict__.items() if isinstance(v, Field)}

    @classmethod
    def _field_name_descriptor_map(cls):
        """
        Returns a dictionary that maps Fields 'Names' to descriptor fields

        >>> class Test(Model):
        ...     first_name = TextField("First Name")
        ...     age = NumberField("Age")
        ...
        >>> Test._field_name_descriptor_map()
        >>> {
        ...     "First Name": <TextField field_name="First Name">,
        ...     "Age": <NumberField field_name="Age">,
        ... }
        """
        return {f.field_name: f for f in cls._attribute_descriptor_map().values()}

    @classmethod
    def _field_name_attribute_map(cls):
        """
        Returns a dictionary that maps Fields 'Names' to the model attribute name:

        >>> class Test(Model):
        ...     first_name = TextField("First Name")
        ...     age = NumberField("Age")
        ...
        >>> Test._field_name_attribute_map()
        >>> {
        ...     "First Name": "first_name"
        ...     "Age": "age"
        ... }
        """
        return {v.field_name: k for k, v in cls._attribute_descriptor_map().items()}

    def __init__(self, **fields):
        # To Store Fields
        self._fields = {}

        # Set descriptors
        for key, value in fields.items():
            if key not in self._attribute_descriptor_map():
                msg = "invalid kwarg '{}'".format(key)
                raise ValueError(msg)
            setattr(self, key, value)

        self.typecast = getattr(self.Meta, "typecast", True)

    @classmethod
    def _validate_class(cls):
        # Verify required Meta attributes were set
        if not getattr(cls.Meta, "base_id", None):
            raise ValueError("Meta.base_id must be defined in model")
        if not getattr(cls.Meta, "table_name", None):
            raise ValueError("Meta.table_name must be defined in model")
        if not getattr(cls.Meta, "api_key", None):
            raise ValueError("Meta.api_key must be defined in model")

        model_attributes = [a for a in cls.__dict__.keys() if not a.startswith("__")]
        overridden = set(model_attributes).intersection(Model.__dict__.keys())
        if overridden:
            raise ValueError(
                "Class {cls} fields clash with existing method: {name}".format(
                    cls=cls.__name__, name=overridden
                )
            )

    @classmethod
    def get_table(cls) -> Table:
        """Return Airtable :class:`~pyairtable.api.Table` class instance"""
        if not hasattr(cls, "_table"):
            cls._table = Table(
                cls.Meta.api_key,  # type: ignore
                cls.Meta.base_id,  # type: ignore
                cls.Meta.table_name,  # type: ignore
                timeout=getattr(cls.Meta, "timeout", None),  # type: ignore
            )
        return cls._table

    def exists(self) -> bool:
        """Returns boolean indicating if instance exists (has 'id' attribute)"""
        return bool(self.id)

    def save(self) -> bool:
        """
        Saves or updates a model.
        If instance has no 'id', it will be created, otherwise updatedself.

        Returns `True` if was created and `False` if it was updated
        """
        table = self.get_table()
        record = self.to_record()
        fields = record["fields"]

        if not self.id:
            record = table.create(fields, typecast=self.typecast)
            did_create = True
        else:
            record = table.update(self.id, fields, replace=True, typecast=self.typecast)
            did_create = False

        self.id = record["id"]
        self.created_time = record["createdTime"]
        return did_create

    def delete(self) -> bool:
        """Deleted record. Must have 'id' field"""
        if not self.id:
            raise ValueError("cannot be deleted because it does not have id")
        table = self.get_table()
        result = table.delete(self.id)
        # Is it even possible go get "deleted" False?
        return result["deleted"]

    @classmethod
    def all(cls, **kwargs) -> List[T]:
        """Returns all records for this model. See :meth:`~pyairtable.api.Api.all`"""
        table = cls.get_table()
        return table.all(**kwargs)

    @classmethod
    def first(cls, **kwargs) -> List[T]:
        """Returns first record for this model. See :meth:`~pyairtable.api.Api.first`"""
        table = cls.get_table()
        return table.first(**kwargs)

    def to_record(self) -> dict:
        """
        Returns a dictionary object as an Airtable record.
        This method converts internal field values into values expected by Airtable.
        e.g. a ``datetime`` value from :class:``DateTimeField`` is converted into an
        ISO 8601 string
        """
        map_ = self._field_name_descriptor_map()
        fields = {k: map_[k].to_record_value(v) for k, v in self._fields.items()}
        return {"id": self.id, "createdTime": self.created_time, "fields": fields}

    @classmethod
    def from_record(cls: Type[T], record: dict) -> T:
        """Create instance from record dictionary"""
        name_attr_map = cls._field_name_attribute_map()
        name_field_map = cls._field_name_descriptor_map()
        try:
            # Convert Column Names into model field names
            # Use field's to_internal to cast into model fields
            kwargs = {
                name_attr_map[k]: name_field_map[k].to_internal_value(v)
                for k, v in record["fields"].items()
            }
        except KeyError as exc:
            raise ValueError("Invalid Field Name: {} for model {}".format(exc, cls))
        instance = cls(**kwargs)
        instance.created_time = record["createdTime"]
        instance.id = record["id"]
        return instance

    @classmethod
    def from_id(cls: Type[T], record_id: str, fetch=True) -> T:
        """
        Create an instance from a `record_id`

        Args:
            record_id: |arg_record_id|

        Keyward Args:
            fetch: If `True`, record will be fetched and fields will be
                updated. If `False`, a new instance is created with the provided `id`,
                but field values are unset. Default is `True`.

        Returns:
            (``Model``): Instance of model
        """
        if fetch:
            table = cls.get_table()
            record = table.get(record_id)
            return cls.from_record(record)
        else:
            instance = cls()
            instance.id = record_id
            return instance

    def fetch(self):
        """Fetches field and resets instance field values from the Airtable record"""
        if not self.id:
            raise ValueError("cannot be fetched because instance does not have an id")

        updated = self.from_id(self.id, fetch=True)
        self._fields = updated._fields
        self.created_time = updated.created_time

    def __repr__(self):
        return "<Model={} {}>".format(self.__class__.__name__, hex(id(self)))

    # TODO - see metadata.py
    # def verify_schema(cls) -> Tuple[bool, dict]:
    #     """verify local airtable models"""

    #     base_list = cls.get_base_list()
    #     base_id_exists = cls.base_id in [b["id"] for b in base_list["bases"]]

    #     if base_id_exists:
    #         base_schema = cls.get_base_schema()
    #         table_schema: dict = next(
    #             (t for t in base_schema["tables"] if t["name"] == cls.table_name), {}
    #         )
    #         table_exists = bool(table_schema)
    #     else:
    #         table_exists = False

    #     if table_exists:
    #         airtable_field_names = [f["name"] for f in table_schema["fields"]]
    #         # Fields.NAME = "|NAME|", Fields.ID = "|ID|") -> ["|NAME|", "|ID|"]
    #         local_field_names = [v for k, v in vars(cls.Fields).items() if k.isupper()]
    #         fields = {n: n in airtable_field_names for n in local_field_names}
    #     else:
    #         fields = {}

    #     in_sync = base_id_exists and table_exists and all(fields.values())
    #     details = {"base": base_id_exists, "table": table_exists, "fields": fields}
    #     return (in_sync, details)



from typing import List
from .api import ApiBase


class Table(ApiBase):
    """
    Represents an Airtable Table. This calss is similar to :class:`~pyairtable.api.Api`,
    except ``base_id`` and ``table_id`` are provided on init instead of provided
    on each method call.

    Usage:
        >>> table = Table('apikey', 'base_id', 'table_name')
        >>> table.all()
    """

    base_id: str
    table_name: str

    def __init__(self, api_key: str, base_id: str, table_name: str, *, timeout=None):
        """
        Args:
            api_key: |arg_api_key|
            base_id: |arg_base_id|
            table_name: |arg_table_name|

        Keyword Args:
            timeout(``Tuple``): |arg_timeout|
        """
        self.base_id = base_id
        self.table_name = table_name
        super().__init__(api_key, timeout=timeout)

    @property
    def table_url(self):
        """Returns the table URL"""
        return super().get_table_url(self.base_id, self.table_name)

    def get_record_url(self, record_id: str):
        """
        Same as :meth:`Api.get_record_url <pyairtable.api.Api.get_record_url>`
        but without ``base_id`` and ``table_name`` arg.
        """
        return super()._get_record_url(self.base_id, self.table_name, record_id)

    def get(self, record_id: str):
        """
        Same as :meth:`Api.get <pyairtable.api.Api.get>`
        but without ``base_id`` and ``table_name`` arg.
        """
        return super()._get_record(self.base_id, self.table_name, record_id)

    def iterate(self, **options):
        """
        Same as :meth:`Api.iterate <pyairtable.api.Api.iterate>`
        but without ``base_id`` and ``table_name`` arg.
        """
        gen = super()._iterate(self.base_id, self.table_name, **options)
        for i in gen:
            yield i

    def first(self, **options):
        """
        Same as :meth:`Api.first <pyairtable.api.Api.first>`
        but without ``base_id`` and ``table_name`` arg.
        """
        return super()._first(self.base_id, self.table_name, **options)

    def all(self, **options):
        """
        Same as :meth:`Api.all <pyairtable.api.Api.all>`
        but without ``base_id`` and ``table_name`` arg.
        """
        return super()._all(self.base_id, self.table_name, **options)

    def create(self, fields: dict, typecast=False):
        """
        Same as :meth:`Api.create <pyairtable.api.Api.create>`
        but without ``base_id`` and ``table_name`` arg.
        """
        return super()._create(self.base_id, self.table_name, fields, typecast=typecast)

    def batch_create(self, records, typecast=False):
        """
        Same as :meth:`Api.batch_create <pyairtable.api.Api.batch_create>`
        but without ``base_id`` and ``table_name`` arg.
        """
        return super()._batch_create(
            self.base_id, self.table_name, records, typecast=typecast
        )

    def update(
        self,
        record_id: str,
        fields: dict,
        replace=False,
        typecast=False,
    ):
        """
        Same as :meth:`Api.update <pyairtable.api.Api.update>`
        but without ``base_id`` and ``table_name`` arg.
        """
        return super()._update(
            self.base_id,
            self.table_name,
            record_id,
            fields,
            replace=replace,
            typecast=typecast,
        )

    def batch_update(self, records: List[dict], replace=False, typecast=False):
        """
        Same as :meth:`Api.batch_update <pyairtable.api.Api.batch_update>`
        but without ``base_id`` and ``table_name`` arg.
        """
        return super()._batch_update(
            self.base_id, self.table_name, records, replace=replace, typecast=typecast
        )

    def delete(self, record_id: str):
        """
        Same as :meth:`Api.delete <pyairtable.api.Api.delete>`
        but without ``base_id`` and ``table_name`` arg.
        """
        return super()._delete(self.base_id, self.table_name, record_id)

    def batch_delete(self, record_ids: List[str]):
        """
        Same as :meth:`Api.batch_delete <pyairtable.api.Api.batch_delete>`
        but without ``base_id`` and ``table_name`` arg.
        """
        return super()._batch_delete(self.base_id, self.table_name, record_ids)

    def __repr__(self) -> str:
        return "<Table base_id={} table_name={}>".format(self.base_id, self.table_name)



from typing import List, Dict, Any
from collections import OrderedDict


class InvalidParamException(ValueError):
    """Raise when invalid parameters are used"""

    def __init__(self, message, *args):
        self.message = message
        super().__init__(message, *args)


def dict_list_to_request_params(param_name: str, values: List[dict]) -> dict:
    """
    Returns dict to be used by request params from dict list

    Expected Airtable Url Params is:
        `?sort[0][field]=FieldOne&sort[0][direction]=asc`

    >>> objects = [
    ...    { "field": "FieldOne", "direction": "asc"},
    ...    { "field": "FieldTwo", "direction": "desc"},
    ... ]
    >>> dict_list_to_request_params("sort", objects)
    {
        "sort[0][field]": "FieldOne",
        "sort[0][direction]: "asc",
        "sort[1][field]": "FieldTwo",
        "sort[1][direction]: "desc",
    }

    """
    param_dict = {}
    for index, dictionary in enumerate(values):
        for key, value in dictionary.items():
            field_name = "{param_name}[{index}][{key}]".format(
                param_name=param_name, index=index, key=key
            )
            param_dict[field_name] = value
    return OrderedDict(sorted(param_dict.items()))


def field_names_to_sorting_dict(field_names: List[str]) -> List[Dict[str, str]]:
    # TODO edge case fields starting with '-'
    """

    >>> field_names_to_sorting_dict(["Name", "-Age"])
    [
        { "field": "FieldOne", "direction": "asc"},
        { "field": "FieldTwo", "direction": "desc"},
    ]
    """
    values = []

    for field_name in field_names:

        if field_name.startswith("-"):
            direction = "desc"
            field_name = field_name[1:]
        else:
            direction = "asc"

        sort_param = {"field": field_name, "direction": direction}
        values.append(sort_param)
    return values


def to_params_dict(param_name: str, value: Any):
    """Returns a dictionary for use in Request 'params'"""
    # TODO: timeZone, userLocale
    if param_name == "max_records":
        return {"maxRecords": value}
    elif param_name == "view":
        return {"view": value}
    elif param_name == "page_size":
        return {"pageSize": value}
    elif param_name == "offset":
        return {"offset": value}
    elif param_name == "formula":
        return {"filterByFormula": value}
    elif param_name == "fields":
        return {"fields[]": value}
    elif param_name == "sort":
        sorting_dict_list = field_names_to_sorting_dict(value)
        return dict_list_to_request_params("sort", sorting_dict_list)
    else:
        msg = "'{0}' is not a supported parameter".format(param_name)
        raise InvalidParamException(msg)



"""
Field are used to define the Airtable column type for your pyAirtable models.

Internally these are implemented as descriptors, this allows us to proxy getting and settings values,
while also providing a type-annotated interface.

>>> from pyairtable.orm import Model, fields
>>> class Contact(Model):
...     name = fields.TextField("Name")
...     is_registered = fields.CheckboxField("Registered")
...
...     class Meta:
...         ...
>>> contact = Contact(name="George", is_registered=True)
>>> assert contact.name == "George"
>>> reveal_type(contact.name)  # -> str
>>> contact.to_record()
{
    "id": recS6qSLw0OCA6Xul",
    "createdTime": "2021-07-14T06:42:37.000Z",
    "fields": {
        "Name": "George",
        "Registered": True,
    }
}


Link Fields
-----------

In addition to standard data type fields, the :class:`LinkField` class
offers a special behaviour that can fetch linked records.

In other words, you can transverse related records through their ``Link Fields``:

>>> from pyairtable.orm import Model, fields
>>> class Company(Model):
...     name = fields.TextField("Name")
...     class Meta:
...         ...
...
>>> class Person(Model):
...     company = fields.LinkField("Company", Company, lazy=False)
...     class Meta:
...         ...
...
>>> contact.from_id("recS6qSLw0OCA6Xul")
>>> contact.company.name # outputs value of Company.name attribute

-----------

"""
import abc
from datetime import date, datetime
from typing import (
    Any,
    TypeVar,
    Type,
    Generic,
    Optional,
    List,
    TYPE_CHECKING,
    Union,
)

from pyairtable import utils

if TYPE_CHECKING:
    from pyairtable.orm import Model  # noqa

T_Linked = TypeVar("T_Linked", bound="Model")


class Field(metaclass=abc.ABCMeta):
    def __init__(self, field_name, validate_type=True) -> None:
        self.field_name = field_name
        self.validate_type = True

    def __set_name__(self, owner, name):
        self.attribute_name = name

    def __get__(self, instance, cls=None):
        # Raise if descriptor is called on class, where instance is None
        if not instance:
            raise RuntimeError("cannot access descriptors on class")

        field_name = self.field_name
        try:
            # Field Field is not yet in dict, return None
            return instance._fields[field_name]
        except (KeyError, AttributeError):
            return None

    def __set__(self, instance, value):
        if not hasattr(instance, "_fields"):
            instance._fields = {}
        if self.validate_type:
            self.valid_or_raise(value)
        instance._fields[self.field_name] = value

    def to_record_value(self, value: Any) -> Any:
        return value

    def to_internal_value(self, value: Any) -> Any:
        return value

    def valid_or_raise(self, value) -> None:
        ...

    def __repr__(self):
        return "<{} field_name='{}'>".format(self.__class__.__name__, self.field_name)


class TextField(Field):
    """Airtable Single Text or Multiline Text Fields. Uses ``str`` to store value"""

    def to_internal_value(self, value: Any) -> str:
        return str(value)

    def valid_or_raise(self, value) -> None:
        if not isinstance(value, str):
            raise ValueError("TextField value must be 'str'")

    def __get__(self, *args, **kwargs) -> Optional[str]:
        return super().__get__(*args, **kwargs)


class EmailField(TextField):
    """Airtable Email field. Uses ``str`` to store value"""

    def valid_or_raise(self, value) -> None:
        if not isinstance(value, str):
            raise ValueError("EmailField value must be 'str'")


class IntegerField(Field):
    """Airtable Number field with Integer Precision. Uses ``int`` to store value"""

    def to_internal_value(self, value: Any) -> int:
        return int(value)

    def valid_or_raise(self, value) -> None:
        if not isinstance(value, int):
            raise ValueError("IntegerField value must be 'int'")

    def __get__(self, *args, **kwargs) -> Optional[int]:
        return super().__get__(*args, **kwargs)


class FloatField(Field):
    """Airtable Number field with Decimal precision. Uses ``float`` to store value"""

    def to_internal_value(self, value: Any) -> float:
        return float(value)

    def valid_or_raise(self, value) -> None:
        if not isinstance(value, float):
            raise ValueError("FloatField value must be 'float'")

    def __get__(self, *args, **kwargs) -> Optional[float]:
        return super().__get__(*args, **kwargs)


class CheckboxField(Field):
    """Airtable Checkbox field. Uses ``bool`` to store value"""

    def to_internal_value(self, value: Any) -> bool:
        return bool(value)

    def valid_or_raise(self, value) -> None:
        if not isinstance(value, bool):
            raise ValueError("CheckboxField value must be 'bool'")

    def __get__(self, *args, **kwargs) -> Optional[bool]:
        return super().__get__(*args, **kwargs)


class DatetimeField(Field):
    """Airtable Datetime field. Uses ``datetime`` to store value"""

    def to_record_value(self, value: datetime) -> str:
        """Airtable expects ISO 8601 string datetime eg. "2014-09-05T12:34:56.000Z" """
        return utils.datetime_to_iso_str(value)

    def to_internal_value(self, value: str) -> datetime:
        """Airtable returns ISO 8601 string datetime eg. "2014-09-05T07:00:00.000Z" """
        return utils.datetime_from_iso_str(value)

    def valid_or_raise(self, value) -> None:
        if not isinstance(value, datetime):
            raise ValueError("DatetimeField value must be 'datetime'")

    def __get__(self, *args, **kwargs) -> Optional[datetime]:
        return super().__get__(*args, **kwargs)


class DateField(Field):
    """Airtable Date field. Uses ``Date`` to store value"""

    def to_record_value(self, value: date) -> str:
        """Airtable expects ISO 8601 date string eg. "2014-09-05"""
        return utils.date_to_iso_str(value)

    def to_internal_value(self, value: str) -> date:
        """Airtable returns ISO 8601 date string eg. "2014-09-05"""
        return utils.date_from_iso_str(value)

    def valid_or_raise(self, value) -> None:
        if not isinstance(value, date):
            raise ValueError("DateField value must be 'date'")

    def __get__(self, *args, **kwargs) -> Optional[datetime]:
        return super().__get__(*args, **kwargs)


class LinkField(Field, Generic[T_Linked]):
    """Airtable Link field. Uses ``List[Model]`` to store value"""

    def __init__(
        self, field_name: str, model: Union[str, Type[T_Linked]], lazy=True
    ) -> None:
        """

        Args:
            field_name: Name of Airtable Column
            model: Model of Linked Type. Must be subtype of :class:`Model`
            lazy: Use `True` to load linked model when looking up attribute. `False`
                will create empty object with only `id` but will not fetch fields.

        Usage:
            >>> TODO
        """

        if isinstance(model, str):
            raise NotImplementedError("path import not implemented")
            # https://github.com/FactoryBoy/factory_boy/blob/37f962720814dff42d7a6a848ccfd200fc7f5ae2/factory/declarations.py#L339
            # model = cast(Type[T_Linked], locate(model))

        assert hasattr(model, "get_table"), "model be subclassed from Model"
        self._model = model
        self._lazy = lazy
        super().__init__(field_name)

    def __get__(self, *args, **kwargs) -> List[T_Linked]:
        return super().__get__(*args, **kwargs)

    def valid_or_raise(self, value) -> None:
        if not hasattr(value, "__iter__"):
            raise TypeError("LinkField value must be iterable")
        for model_instance in value:
            if not isinstance(model_instance, self._model):
                raise ValueError("must be model intance")

    def to_internal_value(self, value: Any) -> List[T_Linked]:
        # If Lazy, create empty from model class and set id
        # If not Lazy, fetch record from pyairtable and create new model instance
        should_fetch = not self._lazy
        linked_models = [
            self._model._linked_cache.get(id_)
            or self._model.from_id(id_, fetch=should_fetch)
            for id_ in value
        ]
        self._model._linked_cache.update({m.id: m for m in linked_models})
        return linked_models

    def to_record_value(self, value: Any) -> List[str]:
        return [v.id for v in value]


"""
- [ ] autoNumber
- [ ] barcode
- [ ] button
- [x] checkbox
- [ ] count
- [ ] createdBy
- [ ] createdTime
- [ ] currency
- [x] date
- [x] dateTime
- [ ] duration
- [x] email
- [ ] externalSyncSource
- [ ] formula
- [ ] lastModifiedBy
- [ ] lastModifiedTime
- [ ] multilineText
- [ ] multipleAttachments
- [ ] multipleCollaborators
- [ ] multipleLookupValues
- [ ] multipleRecordLinks
- [ ] multipleSelects
- [x] number
- [ ] percent
- [ ] phoneNumber
- [ ] rating
- [ ] richText
- [ ] rollup
- [ ] singleCollaborator
- [x] singleLineText
- [ ] singleSelect
- [ ] url
"""



import requests


class AirtableAuth(requests.auth.AuthBase):
    def __init__(self, api_key):
        """
        Authentication used by Airtable Class

        Args:
            api_key (``str``): Airtable API Key.
        """
        self.api_key = api_key

    def __call__(self, request):
        auth_token = {"Authorization": "Bearer {}".format(self.api_key)}
        request.headers.update(auth_token)
        return request



import abc
from functools import lru_cache
import posixpath
from typing import List, Optional
import time
from urllib.parse import quote

import requests

from .auth import AirtableAuth
from .params import to_params_dict


class ApiBase(metaclass=abc.ABCMeta):
    VERSION = "v0"
    API_BASE_URL = "https://api.airtable.com/"
    API_LIMIT = 1.0 / 5  # 5 per second
    API_URL = posixpath.join(API_BASE_URL, VERSION)
    MAX_RECORDS_PER_REQUEST = 10

    def __init__(self, api_key: str, timeout=None):
        session = requests.Session()
        session.auth = AirtableAuth(api_key)
        self.session = session
        self.timeout = timeout

    @lru_cache()
    def get_table_url(self, base_id: str, table_name: str):
        url_safe_table_name = quote(table_name, safe="")
        table_url = posixpath.join(self.API_URL, base_id, url_safe_table_name)
        return table_url

    @lru_cache()
    def _get_record_url(self, base_id: str, table_name: str, record_id):
        """Builds URL with record id"""
        table_url = self.get_table_url(base_id, table_name)
        return posixpath.join(table_url, record_id)

    def _options_to_params(self, **options):
        """
        Process params names or values as needed using filters
        """
        params = {}
        for name, value in options.items():
            params.update(to_params_dict(name, value))
        return params

    def _chunk(self, iterable, chunk_size):
        """Break iterable into chunks"""
        for i in range(0, len(iterable), chunk_size):
            yield iterable[i : i + chunk_size]

    def _build_batch_record_objects(self, records):
        return [{"fields": record} for record in records]

    def _process_response(self, response):
        try:
            response.raise_for_status()
        except requests.exceptions.HTTPError as exc:
            err_msg = str(exc)

            # Attempt to get Error message from response, Issue #16
            try:
                error_dict = response.json()
            except ValueError:
                pass
            else:
                if "error" in error_dict:
                    err_msg += " [Error: {}]".format(error_dict["error"])
            exc.args = (*exc.args, err_msg)
            raise exc
        else:
            return response.json()

    def _request(self, method: str, url: str, params=None, json_data=None):
        response = self.session.request(
            method, url, params=params, json=json_data, timeout=self.timeout
        )
        return self._process_response(response)

    def _get_record(self, base_id: str, table_name: str, record_id: str) -> dict:
        record_url = self._get_record_url(base_id, table_name, record_id)
        return self._request("get", record_url)

    def _iterate(self, base_id: str, table_name: str, **options):
        offset = None
        params = self._options_to_params(**options)
        while True:
            table_url = self.get_table_url(base_id, table_name)
            if offset:
                params.update({"offset": offset})
            data = self._request("get", table_url, params=params)
            records = data.get("records", [])
            time.sleep(self.API_LIMIT)
            yield records
            offset = data.get("offset")
            if not offset:
                break

    def _first(self, base_id: str, table_name: str, **options) -> Optional[dict]:
        for records in self._iterate(
            base_id, table_name, page_size=1, max_records=1, **options
        ):
            for record in records:
                return record
        return None

    def _all(self, base_id: str, table_name: str, **options) -> List[dict]:
        all_records = []

        for records in self._iterate(base_id, table_name, **options):
            all_records.extend(records)
        return all_records

    def _create(self, base_id: str, table_name: str, fields: dict, typecast=False):

        table_url = self.get_table_url(base_id, table_name)
        return self._request(
            "post",
            table_url,
            json_data={"fields": fields, "typecast": typecast},
        )

    def _batch_create(
        self, base_id: str, table_name: str, records: List[dict], typecast=False
    ) -> List[dict]:

        table_url = self.get_table_url(base_id, table_name)
        inserted_records = []
        for chunk in self._chunk(records, self.MAX_RECORDS_PER_REQUEST):
            new_records = self._build_batch_record_objects(chunk)
            response = self._request(
                "post",
                table_url,
                json_data={"records": new_records, "typecast": typecast},
            )
            inserted_records += response["records"]
            time.sleep(self.API_LIMIT)
        return inserted_records

    def _update(
        self,
        base_id: str,
        table_name: str,
        record_id: str,
        fields: dict,
        replace=False,
        typecast=False,
    ) -> List[dict]:
        record_url = self._get_record_url(base_id, table_name, record_id)

        method = "put" if replace else "patch"
        return self._request(
            method, record_url, json_data={"fields": fields, "typecast": typecast}
        )

    def _batch_update(
        self,
        base_id: str,
        table_name: str,
        records: List[dict],
        replace=False,
        typecast=False,
    ):
        updated_records = []
        table_url = self.get_table_url(base_id, table_name)
        method = "put" if replace else "patch"
        for records in self._chunk(records, self.MAX_RECORDS_PER_REQUEST):
            chunk_records = [{"id": x["id"], "fields": x["fields"]} for x in records]
            response = self._request(
                method,
                table_url,
                json_data={"records": chunk_records, "typecast": typecast},
            )
            updated_records += response["records"]

        return updated_records

    def _delete(self, base_id: str, table_name: str, record_id: str):
        record_url = self._get_record_url(base_id, table_name, record_id)
        return self._request("delete", record_url)

    def _batch_delete(
        self, base_id: str, table_name: str, record_ids: List[str]
    ) -> List[dict]:
        deleted_records = []
        table_url = self.get_table_url(base_id, table_name)
        for record_ids in self._chunk(record_ids, self.MAX_RECORDS_PER_REQUEST):
            delete_results = self._request(
                "delete", table_url, params={"records[]": record_ids}
            )
            deleted_records.extend(delete_results["records"])
            time.sleep(self.API_LIMIT)
        return deleted_records


class Api(ApiBase):
    """
    Represents an Airtable Api.

    The Api Key is provided on init and ``base_id`` and ``table_id``
    can be provided on each method call.

    If you are only operating on one Table, or one Base, consider using
    :class:`Base` or :class:`Table`.

    Usage:
        >>> api = Api('apikey')
        >>> api.all('base_id', 'table_name')
    """

    def __init__(self, api_key: str, timeout=None):
        """

        Args:
            api_key: |arg_api_key|

        Keyword Args:
            timeout(``Tuple``): |arg_timeout|

        """
        super().__init__(api_key, timeout=timeout)

    def get_record_url(self, base_id: str, table_name: str, record_id: str):
        """
        Returns a url for the provided record

        Args:
            base_id: |arg_base_id|
            table_name: |arg_table_name|

        """
        return super()._get_record_url(base_id, table_name, record_id)

    def get(self, base_id: str, table_name: str, record_id: str):
        """
        Retrieves a record by its id

        >>> record = api.get('base_id', 'table_name', 'recwPQIfs4wKPyc9D')

        Args:
            base_id: |arg_base_id|
            table_name: |arg_table_name|
            record_id: |arg_record_id|

        Returns:
            record: Record
        """
        return super()._get_record(base_id, table_name, record_id)

    def iterate(self, base_id: str, table_name: str, **options):
        """
        Record Retriever Iterator

        Returns iterator with lists in batches according to pageSize.
        To get all records at once use :meth:`all`

        >>> for page in api.iterate('base_id', 'table_name'):
        ...     for record in page:
        ...         print(record)
        {"id": ... }
        ...

        Args:
            base_id: |arg_base_id|
            table_name: |arg_table_name|

        Keyword Args:
            view: |kwarg_view|
            page_size: |kwarg_page_size|
            max_records: |kwarg_max_records|
            fields: |kwarg_fields|
            sort: |kwarg_sort|
            formula: |kwarg_formula|

        Returns:
            iterator: Record Iterator, grouped by page size

        """
        gen = super()._iterate(base_id, table_name, **options)
        for i in gen:
            yield i

    def first(self, base_id: str, table_name: str, **options):
        """
        Retrieves the first found record or ``None`` if no records are returned.

        This is similar to :meth:`~pyairtable.api.api.Api.all`, except it
        it sets ``page_size`` and ``max_records`` to ``1``.

        Args:
            base_id: |arg_base_id|
            table_name: |arg_table_name|

        Keyword Args:
            view: |kwarg_view|
            fields: |kwarg_fields|
            sort: |kwarg_sort|
            formula: |kwarg_formula|
        """
        return super()._first(base_id, table_name, **options)

    def all(self, base_id: str, table_name: str, **options):
        """
        Retrieves all records repetitively and returns a single list.

        >>> api.all('base_id', 'table_name', view='MyView', fields=['ColA', '-ColB'])
        [{'fields': ... }, ...]
        >>> api.all('base_id', 'table_name', maxRecords=50)
        [{'fields': ... }, ...]

        Args:
            base_id: |arg_base_id|
            table_name: |arg_table_name|

        Keyword Args:
            view: |kwarg_view|
            page_size: |kwarg_page_size|
            max_records: |kwarg_max_records|
            fields: |kwarg_fields|
            sort: |kwarg_sort|
            formula: |kwarg_formula|

        Returns:
            records (``list``): List of Records

        >>> records = all(maxRecords=3, view='All')

        """
        return super()._all(base_id, table_name, **options)

    def create(self, base_id: str, table_name: str, fields: dict, typecast=False):
        """
        Creates a new record

        >>> record = {'Name': 'John'}
        >>> api.create('base_id', 'table_name', record)

        Args:
            base_id: |arg_base_id|
            table_name: |arg_table_name|
            fields(``dict``): Fields to insert.
                Must be dictionary with Column names as Key.

        Keyword Args:
            typecast: |kwarg_typecast|

        Returns:
            record (``dict``): Inserted record

        """
        return super()._create(base_id, table_name, fields, typecast=typecast)

    def batch_create(self, base_id: str, table_name: str, records, typecast=False):
        """
        Breaks records into chunks of 10 and inserts them in batches.
        Follows the set API rate.
        To change the rate limit you can change ``API_LIMIT = 0.2``
        (5 per second)

        >>> records = [{'Name': 'John'}, {'Name': 'Marc'}]
        >>> api.batch_insert('base_id', 'table_name', records)

        Args:
            base_id: |arg_base_id|
            table_name: |arg_table_name|
            records(``List[dict]``): List of dictionaries representing
                records to be created.

        Keyword Args:
            typecast: |kwarg_typecast|

        Returns:
            records (``list``): list of added records
        """
        return super()._batch_create(base_id, table_name, records, typecast=typecast)

    def update(
        self,
        base_id: str,
        table_name: str,
        record_id: str,
        fields: dict,
        replace=False,
        typecast=False,
    ):
        """
        Updates a record by its record id.
        Only Fields passed are updated, the rest are left as is.

        >>> table.update('recwPQIfs4wKPyc9D', {"Age": 21})
        {id:'recwPQIfs4wKPyc9D', fields': {"First Name": "John", "Age": 21}}
        >>> table.update('recwPQIfs4wKPyc9D', {"Age": 21}, replace=True)
        {id:'recwPQIfs4wKPyc9D', fields': {"Age": 21}}

        Args:
            base_id: |arg_base_id|
            table_name: |arg_table_name|
            record_id: |arg_record_id|
            fields(``dict``): Fields to update.
                Must be dictionary with Column names as Key

        Keyword Args:
            replace (``bool``, optional): If ``True``, record is replaced in its entirety
                by provided fields - eg. if a field is not included its value will
                bet set to null. If False, only provided fields are updated.
                Default is ``False``.
            typecast: |kwarg_typecast|

        Returns:
            record (``dict``): Updated record
        """

        return super()._update(
            base_id,
            table_name,
            record_id,
            fields,
            replace=replace,
            typecast=typecast,
        )

    def batch_update(
        self,
        base_id: str,
        table_name: str,
        records: List[dict],
        replace=False,
        typecast=False,
    ):
        """
        Updates a records by their record id's in batch.

        Args:
            base_id: |arg_base_id|
            table_name: |arg_table_name|
            records(``list``): List of dict: [{"id": record_id, "field": fields_to_update_dict}]

        Keyword Args:
            replace (``bool``, optional): If ``True``, record is replaced in its entirety
                by provided fields - eg. if a field is not included its value will
                bet set to null. If False, only provided fields are updated.
                Default is ``False``.
            typecast: |kwarg_typecast|

        Returns:
            records(``list``): list of updated records
        """
        return super()._batch_update(
            base_id, table_name, records, replace=replace, typecast=typecast
        )

    def delete(self, base_id: str, table_name: str, record_id: str):
        """
        Deletes a record by its id

        >>> record = api.match('base_id', 'table_name', 'Employee Id', 'DD13332454')
        >>> api.delete('base_id', 'table_name', record['id'])

        Args:
            base_id: |arg_base_id|
            table_name: |arg_table_name|
            record_id: |arg_record_id|

        Returns:
            record (``dict``): Deleted Record
        """
        return super()._delete(base_id, table_name, record_id)

    def batch_delete(self, base_id: str, table_name: str, record_ids: List[str]):
        """
        Breaks records into batches of 10 and deletes in batches, following set
        API Rate Limit (5/sec).
        To change the rate limit set value of ``API_LIMIT`` to
        the time in seconds it should sleep before calling the function again.

        >>> record_ids = ['recwPQIfs4wKPyc9D', 'recwDxIfs3wDPyc3F']
        >>> api.batch_delete('base_id', 'table_name', records_ids)

        Args:
            base_id: |arg_base_id|
            table_name: |arg_table_name|
            record_ids(``list``): Record Ids to delete

        Returns:
            records(``list``): list of records deleted

        """
        return super()._batch_delete(base_id, table_name, record_ids)

    def __repr__(self) -> str:
        return "<Airtable Api>"



from datetime import datetime, date
from typing import Union


def datetime_to_iso_str(value: datetime) -> str:
    """
    Converts ``datetime`` object into Airtable compatible ISO 8601 string
    e.g. "2014-09-05T12:34:56.000Z"

    Args:
        value: datetime object
    """
    return value.isoformat(timespec="milliseconds") + "Z"


def datetime_from_iso_str(value: str) -> datetime:
    """
    Converts ISO 8601 datetime string into a ``datetime`` object.
    Expected format is "2014-09-05T07:00:00.000Z"

    Args:
        value: datetime string e.g. "2014-09-05T07:00:00.000Z"
    """
    return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")


def date_to_iso_str(value: Union[date, datetime]) -> str:
    """
    Converts ``date`` or ``datetime`` object into Airtable compatible ISO 8601 string
    e.g. "2014-09-05"

    Args:
        value: date or datetime object
    """
    return value.strftime("%Y-%m-%d")


def date_from_iso_str(value: str) -> date:
    """
    Converts ISO 8601 date string into a ``date`` object.
    Expected format is  "2014-09-05"

    Args:
        value: date string e.g. "2014-09-05"
    """
    return datetime.strptime(value, "%Y-%m-%d").date()



import posixpath
from typing import Union, Optional
from pyairtable.api import Api, Base, Table


def get_api_bases(api: Union[Api, Base]) -> dict:
    """
    Return list of Bases from an Api or Base instance.
    For More Details `Metadata Api Documentation <https://airtable.com/api/meta>`_

    Args:
        api: :class:`Api` or :class:`Base` instance

    Usage:
        >>> table.get_bases()
            {
                "bases": [
                    {
                        "id": "appY3WxIBCdKPDdIa",
                        "name": "Apartment Hunting",
                        "permissionLevel": "create"
                    },
                    {
                        "id": "appSW9R5uCNmRmfl6",
                        "name": "Project Tracker",
                        "permissionLevel": "edit"
                    }
                ]
            }
    """
    base_list_url = posixpath.join(api.API_URL, "meta", "bases")
    return api._request("get", base_list_url)


def get_base_schema(base: Union[Base, Table]) -> dict:
    """
    Returns Schema of a Base
    For More Details `Metadata Api Documentation <https://airtable.com/api/meta>`_

    Args:
        base: :class:`Base` or :class:`Table` instance

    Usage:
        >>> get_base_schema(base)
            {
                "tables": [
                    {
                        "id": "tbltp8DGLhqbUmjK1",
                        "name": "Apartments",
                        "primaryFieldId": "fld1VnoyuotSTyxW1",
                        "fields": [
                            {
                                "id": "fld1VnoyuotSTyxW1",
                                "name": "Name",
                                "type": "singleLineText"
                            },
                            {
                                "id": "fldoaIqdn5szURHpw",
                                "name": "Pictures",
                                "type": "multipleAttachment"
                            },
                            {
                                "id": "fldumZe00w09RYTW6",
                                "name": "District",
                                "type": "multipleRecordLinks"
                            }
                        ],
                        "views": [
                            {
                                "id": "viwQpsuEDqHFqegkp",
                                "name": "Grid view",
                                "type": "grid"
                            }
                        ]
                    }
                ]
            }
    """
    base_schema_url = posixpath.join(
        base.API_URL, "meta", "bases", base.base_id, "tables"
    )
    return base._request("get", base_schema_url)


def get_table_schema(table: Table) -> Optional[dict]:
    """
    Returns the specific table schema record provided by base schema list

    Args:
        table: :class:`Table` instance

    Usage:
        >>> get_table_schema(table)
        {
            "id": "tbltp8DGLhqbUmjK1",
            "name": "Apartments",
            "primaryFieldId": "fld1VnoyuotSTyxW1",
            "fields": [
                {
                    "id": "fld1VnoyuotSTyxW1",
                    "name": "Name",
                    "type": "singleLineText"
                }
            ],
            "views": [
                {
                    "id": "viwQpsuEDqHFqegkp",
                    "name": "Grid view",
                    "type": "grid"
                }
            ]
        }
    """
    base_schema = get_base_schema(table)
    for table_record in base_schema.get("tables", {}):
        if table.table_name == table_record["name"]:
            return table_record
    return None



from typing import List

from .api import ApiBase


class Base(ApiBase):
    """
    Represents an Airtable Base. This calss is similar to :class:`~pyairtable.api.Api`,
    except ``base_id`` is provided on init instead of provided on each method call.

    Usage:
        >>> base = Base('apikey', 'base_id')
        >>> base.all()
    """

    base_id: str

    def __init__(self, api_key: str, base_id: str, timeout=None):
        """
        Args:
            api_key: |arg_api_key|
            base_id: |arg_base_id|

        Keyword Args:
            timeout(``Tuple``): |arg_timeout|
        """
        self.base_id = base_id
        super().__init__(api_key, timeout=timeout)

    def get_record_url(self, table_name: str, record_id: str):
        """
        Same as :meth:`Api.get_record_url <pyairtable.api.Api.get_record_url>`
        but without ``base_id`` arg.
        """
        return super()._get_record_url(self.base_id, table_name, record_id)

    def get(self, table_name: str, record_id: str):
        """
        Same as :meth:`Api.get <pyairtable.api.Api.get>`
        but without ``base_id`` arg.
        """
        return super()._get_record(self.base_id, table_name, record_id)

    def iterate(self, table_name: str, **options):
        """
        Same as :meth:`Api.iterate <pyairtable.api.Api.iterate>`
        but without ``base_id`` arg.
        """
        gen = super()._iterate(self.base_id, table_name, **options)
        for i in gen:
            yield i

    def first(self, table_name: str, **options):
        """
        Same as :meth:`Api.first <pyairtable.api.Api.first>`
        but without ``base_id`` arg.
        """
        return super()._first(self.base_id, table_name, **options)

    def all(self, table_name: str, **options):
        """
        Same as :meth:`Api.all <pyairtable.api.Api.all>`
        but without ``base_id`` arg.
        """
        return super()._all(self.base_id, table_name, **options)

    def create(self, table_name: str, fields: dict, typecast=False):
        """
        Same as :meth:`Api.create <pyairtable.api.Api.create>`
        but without ``base_id`` arg.
        """
        return super()._create(self.base_id, table_name, fields, typecast=typecast)

    def batch_create(self, table_name: str, records, typecast=False):
        """
        Same as :meth:`Api.batch_create <pyairtable.api.Api.batch_create>`
        but without ``base_id`` arg.
        """
        return super()._batch_create(
            self.base_id, table_name, records, typecast=typecast
        )

    def update(
        self,
        table_name: str,
        record_id: str,
        fields: dict,
        replace=False,
        typecast=False,
    ):
        """
        Same as :meth:`Api.update <pyairtable.api.Api.update>`
        but without ``base_id`` arg.
        """
        return super()._update(
            self.base_id,
            table_name,
            record_id,
            fields,
            replace=replace,
            typecast=typecast,
        )

    def batch_update(
        self, table_name: str, records: List[dict], replace=False, typecast=False
    ):
        """
        Same as :meth:`Api.batch_update <pyairtable.api.Api.batch_update>`
        but without ``base_id`` arg.
        """
        return super()._batch_update(
            self.base_id, table_name, records, replace=replace, typecast=typecast
        )

    def delete(self, table_name: str, record_id: str):
        """
        Same as :meth:`Api.delete <pyairtable.api.Api.delete>`
        but without ``base_id`` arg.
        """
        return super()._delete(self.base_id, table_name, record_id)

    def batch_delete(self, table_name: str, record_ids: List[str]):
        """
        Same as :meth:`Api.batch_delete <pyairtable.api.Api.batch_delete>`
        but without ``base_id`` arg.
        """
        return super()._batch_delete(self.base_id, table_name, record_ids)

    def __repr__(self) -> str:
        return "<Airtable Base id={}>".format(self.base_id)
